import math
from functools import partial
from typing import Dict, List, Literal, Optional, Sequence

import torch

from mppq.executor.base import RuntimeHook
from mppq.ir.base.opdef import Operation
from mppq.logger import warning
from mppq.quantization.measure.cosine import torch_cosine_similarity
from mppq.quantization.measure.norm import torch_mean_square_error, torch_snr_error
from mppq.utils.fetch import batch_random_fetch, tensor_random_fetch


class OutputRecorder(RuntimeHook):
    """Record model output in the runtime hook."""

    def __init__(self, operation: Operation, fetches: int = 4096) -> None:
        self.fetched: Optional[torch.Tensor] = None
        self.fetches = fetches
        super().__init__(operation)

    def post_forward_hook(self, outputs: Sequence[torch.Tensor], **kwargs):
        output_tensor = outputs[0]
        assert isinstance(
            output_tensor, torch.Tensor
        ), "Output of monitoring operation is not a torch.Tensor"
        self.fetched = batch_random_fetch(
            output_tensor, seed=10086, fetches_per_batch=self.fetches
        ).to("cpu")
        return super().post_forward_hook(outputs, **kwargs)

    def pop(self) -> torch.Tensor:
        assert self.fetched is not None
        fetched = self.fetched
        self.fetched = None
        return fetched


class DetailedRecorder(RuntimeHook):
    def __init__(self, operation: Operation, fetches: int = 1024) -> None:
        self.fetches = fetches
        self.i_storage = [[] for _ in range(operation.num_of_input)]
        self.o_storage = [[] for _ in range(operation.num_of_output)]
        super().__init__(operation)

    def pre_forward_hook(self, inputs: Sequence[torch.Tensor], **kwargs):
        for i, v in enumerate(inputs):
            vv = tensor_random_fetch(v, seed=10086, num_of_fetches=self.fetches)
            self.i_storage[i].append(vv.to("cpu"))
        return super().pre_forward_hook(inputs, **kwargs)

    def post_forward_hook(self, outputs: Sequence[torch.Tensor], **kwargs):
        for i, v in enumerate(outputs):
            vv = tensor_random_fetch(v, seed=10086, num_of_fetches=self.fetches)
            self.o_storage[i].append(vv.to("cpu"))
        return super().post_forward_hook(outputs, **kwargs)

    def clear(self):
        self.i_storage = [[] for _ in range(self._hook_to.num_of_input)]
        self.o_storage = [[] for _ in range(self._hook_to.num_of_output)]


class MeasureRecorder:
    """Helper class for collecting data."""

    def __init__(self, measurement: str = "cosine", reduce: str = "mean") -> None:
        self.num_of_elements: int = 0
        self.measure: float = 0
        if reduce not in {"mean", "max"}:
            raise ValueError(
                "PPQ MeasureRecorder Only support reduce by mean or max, "
                f"however {reduce} was given."
            )

        if str(measurement).lower() == "cosine":
            measure_fn = partial(torch_cosine_similarity, reduction=reduce)
        elif str(measurement).lower() == "mse":
            measure_fn = partial(torch_mean_square_error, reduction=reduce)
        elif str(measurement).lower() == "snr":
            measure_fn = partial(torch_snr_error, reduction=reduce)
        else:
            raise ValueError(
                "Unsupported measurement detected. PPQ only support mse, "
                f"snr and consine now, while {measurement} was given."
            )

        self.measure_fn = measure_fn
        self.reduce = reduce

    def update(self, y_pred: torch.Tensor, y_real: torch.Tensor):
        elements = y_pred.shape[0]
        if elements != y_real.shape[0]:
            raise ValueError(
                "Can not update measurement, cause your input data do not share a "
                f"same batchsize. Shape of y_pred {y_pred.shape} - against shape of "
                f"y_real {y_real.shape}"
            )
        result = self.measure_fn(y_pred=y_pred, y_real=y_real).item()

        if self.reduce == "mean":
            self.measure = self.measure * self.num_of_elements + result * elements
            self.num_of_elements += elements
            self.measure /= self.num_of_elements

        if self.reduce == "max":
            self.measure = max(self.measure, result)
            self.num_of_elements += elements


class MeasurePrinter:
    """Helper class for print top-k record."""

    def __init__(
        self,
        data: Dict[str, float],
        measure: str,
        label: str = "Layer",
        k: Optional[int] = None,
        order: Optional[Literal["large_to_small", "small_to_large"]] = "large_to_small",
        percentage: bool = False,
    ) -> None:

        if order not in {"large_to_small", "small_to_large", None}:
            raise ValueError(
                'Parameter "order" can only be "large_to_small" or "small_to_large"'
            )
        self.collection = [(name, value) for name, value in data.items()]
        if order is not None:
            self.collection = sorted(self.collection, key=lambda x: x[1])
            if order == "large_to_small":
                self.collection = self.collection[::-1]
        if k is not None:
            self.collection = self.collection[:k]

        if order is None:
            sorted_collection = sorted(self.collection, key=lambda x: x[1])
            largest_element, smallest_element = (
                sorted_collection[-1][1],
                sorted_collection[0][1],
            )
        elif order == "large_to_small":
            largest_element, smallest_element = (
                self.collection[0][1],
                self.collection[-1][1],
            )
        else:  # order == "small_to_large"
            largest_element, smallest_element = (
                self.collection[-1][1],
                self.collection[0][1],
            )
        self.normalized_by = largest_element - smallest_element
        self.min = smallest_element

        max_name_length = len(label)
        for name, _ in self.collection:
            max_name_length = max(len(name), max_name_length)
        self.max_name_length = max_name_length
        self.measure_str = measure
        self.label = label
        self.percentage = percentage

    def print(self, max_blocks: int = 20, ansi: bool = False):
        """Pretty print of measured data."""

        print(f"{self.label:<{self.max_name_length}} | {self.measure_str}")
        ch = "=" if ansi else "█"  # \u2588
        for name, value in self.collection:
            normalized_value = (value - self.min) / (self.normalized_by + 1e-7)
            if math.isnan(value):
                warning("MeasurePrinter found an NaN value in your data.")
                normalized_value = 0
            num_of_blocks = round(normalized_value * max_blocks)
            # align the text
            num_of_blocks = max(min(max_blocks - 1, num_of_blocks), 1)

            msg = f"{name:<{self.max_name_length}} "
            msg += f"{'|':{ch}<{num_of_blocks}}{'|':>{max_blocks - num_of_blocks}} "
            if self.percentage:
                msg += f"{value * 100:.3f}%"
            else:
                msg += f"{value:.4f}"
            print(msg)



import numpy as np
import pytest
import torch

from mppq.data import (
    DataType,
    convert_any_to_numpy,
    convert_any_to_python_primary_type,
    convert_any_to_tensor,
)


# pylint: disable=missing-function-docstring
@pytest.mark.parametrize(
    "dtype",
    [
        np.bool,
        np.uint8,
        np.int8,
        np.int16,
        np.int32,
        np.int64,
        np.float16,
        np.float32,
        np.float64,
        np.dtype(np.bool),
        np.dtype(np.uint8),
        np.dtype(np.int8),
        np.dtype(np.int16),
        np.dtype(np.int32),
        np.dtype(np.int64),
        np.dtype(np.float16),
        np.dtype(np.float32),
        np.dtype(np.float64),
    ],
)
def test_data_type_from_numpy(dtype):
    mytype = DataType.from_numpy(dtype)
    assert DataType.to_numpy(mytype) == dtype


@pytest.mark.parametrize(
    "dtype",
    [
        torch.bool,
        torch.uint8,
        torch.int8,
        torch.int16,
        torch.int32,
        torch.int64,
        torch.float16,
        torch.float32,
        torch.float64,
        torch.float8_e5m2,
    ],
)
def test_data_type_from_torch(dtype):
    mytype = DataType.from_torch(dtype)
    assert DataType.to_torch(mytype) == dtype


@pytest.mark.parametrize(
    ("value", "accept_none", "expected"),
    [
        (1, False, np.array([1])),
        ([1, 2, 3], False, np.array([1, 2, 3])),
        ((1, 2, 3), False, np.array((1, 2, 3))),
        ("123", False, np.array("123")),
        (np.array([1, 2, 3]), False, np.array([1, 2, 3])),
        (np.empty(0), False, ValueError),
        (np.empty(0), True, None),
        (np.array(1), False, np.array(1)),
        (torch.tensor([]), False, ValueError),
        (torch.tensor([]), True, None),
        (torch.tensor(1), False, np.array([1])),
        (torch.tensor([1, 2, 3]), False, np.array([1, 2, 3])),
    ],
)
def test_convert_any_to_numpy(value, accept_none, expected):
    if isinstance(expected, type(object)) and issubclass(expected, Exception):
        with pytest.raises(expected):
            convert_any_to_numpy(value, accept_none=accept_none)
    else:
        np_value = convert_any_to_numpy(value, accept_none=accept_none)
        assert np.all(np_value == expected)


@pytest.mark.parametrize(
    ("value", "expected"),
    [
        (1, torch.tensor([1])),
        ([1, 2, 3], torch.tensor([1, 2, 3])),
        ((1, 2, 3), torch.tensor((1, 2, 3))),
        ("123", TypeError),
        (np.array([1, 2, 3]), torch.tensor([1, 2, 3])),
        (np.empty(0), torch.tensor([])),
        (np.empty(0), torch.tensor([])),
        (np.array(1), torch.tensor(1)),
        (torch.tensor([]), torch.tensor([])),
        (torch.tensor([]), torch.tensor([])),
        (torch.tensor(1), torch.tensor([1])),
        (torch.tensor([1, 2, 3]), torch.tensor([1, 2, 3])),
    ],
)
def test_convert_any_to_tensor(value, expected):
    if isinstance(expected, type(object)) and issubclass(expected, Exception):
        with pytest.raises(expected):
            convert_any_to_tensor(value)
    else:
        tensor_value = convert_any_to_tensor(value)
        assert torch.all(tensor_value == expected)


@pytest.mark.parametrize(
    ("value", "accept_none", "expected"),
    [
        (np.array([1, 2, 3]), False, [1, 2, 3]),
        (torch.tensor([1, 2, 3]), False, [1, 2, 3]),
        (np.array([1]), False, 1),
        (torch.tensor([1]), False, 1),
        (np.empty(0), False, ValueError),
        (np.empty(0), True, None),
        (torch.empty(0), False, ValueError),
        (torch.empty(0), True, None),
        (1, False, 1),
        (None, False, ValueError),
        (None, True, None),
        ([1, 2, 3], False, [1, 2, 3]),
    ],
)
def test_convert_any_to_python_primary_type(value, accept_none, expected):
    if isinstance(expected, type(object)) and issubclass(expected, Exception):
        with pytest.raises(expected):
            convert_any_to_python_primary_type(value, accept_none=accept_none)
    else:
        python_value = convert_any_to_python_primary_type(
            value, accept_none=accept_none
        )
        assert python_value == expected



import torch

from mppq.data import convert_any_to_tensor
from mppq.frontend.onnx.onnxruntime_exporter import ONNXRUNTIMExporter
from mppq.ir.base.graph import BaseGraph
from mppq.ir.base.opdef import Operation, Variable
from mppq.quant import QuantizationProperty, TensorQuantizationConfig
from mppq.utils.round import ppq_tensor_round


class OpenvinoExporter(ONNXRUNTIMExporter):
    """Openvino 喜欢所有 QuantizeLinear 上面都带 axis 属性，所以这里需要单独处理一下"""

    def insert_quantize_node(
        self,
        graph: BaseGraph,
        var: Variable,
        config: TensorQuantizationConfig,
        op: Operation,
    ) -> Operation:
        """
        Insert a Quantize Node on given variable, according to given TensorQuantizationConfig.
        There is two basic type of Quantize Node: QuantizeLinear and QuantizeFloating.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            # Following code will export Linear Quantization Config
            # That is for FP32 -> INT
            offset_dtype, value_type = self.infer_qtype(config)
            scale = convert_any_to_tensor(config.scale.clone(), dtype=torch.float32)
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

            created = graph.create_operation(op_type="QuantizeLinear", attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes["axis"] = config.channel_axis
            else:
                created.attributes["axis"] = 0

            if var in op.inputs:
                graph.insert_op_before(created, op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(created, op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(
                    f"Unexpected Error in Exporting Op {op.name}({op.type})."
                )

            graph.create_variable(
                name=None, value=scale, is_parameter=True, dest_ops=[created]
            )
            graph.create_variable(
                name=None, value=offset, is_parameter=True, dest_ops=[created]
            )

            created.outputs[0].dtype = value_type
            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape
            return created

        elif config.policy.has_property(QuantizationProperty.FLOATING):
            # Following code will export Linear Quantization Config
            # That is for FP32 -> FP8
            scale = convert_any_to_tensor(config.scale.clone(), dtype=torch.float32)
            offset = convert_any_to_tensor(config.offset.clone(), dtype=torch.float32)

            created = graph.create_operation(
                op_type="QuantizeFloating",
                attributes={
                    "min": config.quant_min,
                    "max": config.quant_max,
                    "exponent": config.exponent_bits,
                    "mantissa": config.mantissa_bits,
                },
            )

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes["axis"] = config.channel_axis
            else:
                created.attributes["axis"] = None

            if var in op.inputs:
                graph.insert_op_before(created, op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(created, op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(
                    f"Unexpected Error in Exporting Op {op.name}({op.type})."
                )

            graph.create_variable(
                name=None, value=scale, is_parameter=True, dest_ops=[created]
            )
            graph.create_variable(
                name=None, value=offset, is_parameter=True, dest_ops=[created]
            )

            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape
            return created

        else:
            raise TypeError(
                f"PPQ Can not export quantization information with variable {var.name}, "
                "Unexpected Quantization property."
            )

    def insert_dequantize_node(
        self,
        graph: BaseGraph,
        var: Variable,
        config: TensorQuantizationConfig,
        op: Operation,
    ) -> Operation:
        """
        Insert a DeQuantize Node on given variable, according to given TensorQuantizationConfig.
        There is two basic type of DeQuantize Node: DeQuantizeLinear and DeQuantizeFloating.
        """
        if config.policy.has_property(QuantizationProperty.LINEAR):
            offset_dtype, value_type = self.infer_qtype(config)
            scale = convert_any_to_tensor(config.scale.clone(), dtype=torch.float32)
            offset = ppq_tensor_round(config.offset.clone()).type(offset_dtype)

            created = graph.create_operation(op_type="DequantizeLinear", attributes={})
            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes["axis"] = config.channel_axis
            else:
                created.attributes["axis"] = 0

            if var in op.inputs:
                graph.insert_op_before(created, op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(created, op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(
                    f"Unexpected Error in Exporting Op {op.name}({op.type})."
                )

            graph.create_variable(
                name=None, value=scale, is_parameter=True, dest_ops=[created]
            )
            graph.create_variable(
                name=None, value=offset, is_parameter=True, dest_ops=[created]
            )

            created.inputs[0].dtype = value_type
            created.inputs[0].shape = var.shape
            created.outputs[0].shape = var.shape
            created.outputs[0].dtype = torch.float32

            return created

        elif config.policy.has_property(QuantizationProperty.FLOATING):
            scale = convert_any_to_tensor(config.scale.clone(), dtype=torch.float32)
            offset = convert_any_to_tensor(config.offset.clone(), dtype=torch.float32)

            created = graph.create_operation(
                op_type="DequantizeFloating",
                attributes={
                    "min": config.quant_min,
                    "max": config.quant_max,
                    "exponent": config.exponent_bits,
                    "mantissa": config.mantissa_bits,
                },
            )

            if config.policy.has_property(QuantizationProperty.PER_CHANNEL):
                created.attributes["axis"] = config.channel_axis
            else:
                created.attributes["axis"] = None

            if var in op.inputs:
                graph.insert_op_before(created, op, input_idx=op.inputs.index(var))
            elif var in op.outputs:
                graph.insert_op_after(created, op, output_idx=op.outputs.index(var))
            else:
                raise ValueError(
                    f"Unexpected Error in Exporting Op {op.name}({op.type})."
                )

            graph.create_variable(
                name=None, value=scale, is_parameter=True, dest_ops=[created]
            )
            graph.create_variable(
                name=None, value=offset, is_parameter=True, dest_ops=[created]
            )

            created.outputs[0].shape = var.shape
            created.inputs[0].shape = var.shape

            return created

        else:

            raise TypeError(
                f"PPQ Can not export quantization information with variable {var.name}, "
                "Unexpected Quantization property."
            )



from abc import ABCMeta, abstractmethod
from typing import Any, List, Optional, TypeVar

from mppq.ir.base.command import GraphCommand, GraphCommandType
from mppq.ir.base.graph import BaseGraph

T = TypeVar("T", BaseGraph, "GraphCommandProcessor")


class GraphCommandProcessor(metaclass=ABCMeta):
    r"""用于处理图上相关操作的抽象基类.

    我们使用指令-责任链模式处理 PPQ 计算图的相关操作，具体来说：

    所有图上相关操作都由一个 GraphCommand 对象进行封装，
    这一对象封装了操作的类型和必要参数

    同时我们设计了 GraphCommandProcessor 类用于接收并处理对应的 GraphCommand

    GraphCommandProcessor 被设计为责任链模式，当接收到无法处理的 GraphCommand 时
    将把无法识别 GraphCommand 传递给责任链上的下一任 GraphCommandProcessor
    直到 GraphCommand 被处理，或抛出无法处理的异常

    当你实现一个新的 GraphCommandProcessor 时，需要实现其中的方法 _acceptable_command_types，
    该方法返回了所有可以被识别的 GraphCommand 类型，同时在 _process 的逻辑中对 GraphCommand 的请求进行处理

    这两个方法被设计为私有的，这意味着你不能单独访问责任链中的独立 GraphCommandProcessor，
    只能够通过责任链的方式发起请求

    如果在责任链中有多个可以处理同一类请求的 GraphCommandProcessor，
    只有最先接触到 GraphCommand 的 GraphCommandProcessor将会被调用

    GraphCommandProcessor 将按照自定义逻辑解析 GraphCommand，
    在 BaseGraph 做出相应处理并给出结果，实现方法请参考 RunnableGraph

    Args:
        graph_or_processor (BaseGraph, Callable): 被处理的图对象，可以是 BaseGraph 或者
            GraphCommandProcessor. 如果是 GraphCommandProcessor 对象，则自动将 self 与 graph 链接成链
    """

    def __init__(self, graph_or_processor: T) -> None:
        self._next_command_processor: Optional[GraphCommandProcessor] = None
        if isinstance(graph_or_processor, GraphCommandProcessor):
            self._next_command_processor = graph_or_processor
            self._graph = graph_or_processor._graph
        else:
            self._graph = graph_or_processor

    @property
    @abstractmethod
    def _acceptable_command_types(self) -> List[GraphCommandType]:
        """Subclass of GraphCommandProcessor must give an implementation of
        this function.

            Return all acceptable GraphCommandTypes in a list as result.
            something like:
                return [
                    GraphCommandType.DEPLOY_TO_CPU,
                    GraphCommandType.DEPLOY_TO_CUDA,
                    GraphCommandType.DEPLOY_TO_NUMPY
                ]

        Returns:
            List[GraphCommandType]: all acceptable GraphCommandTypes
        """
        raise NotImplementedError

    def __call__(self, command: GraphCommand) -> Any:
        """Invoking interface of GraphCommandProcessor responsibility chain.
        All processors within the chain shall be invoked by this function one
        be one, until there is a processor claim to accept input command
        object, the entire processing of responsibility chain ends then.

        invoke a GraphCommandProcessor chain like that:

            _ = GraphCommandProcessor(graph, next_command_processor=None)
            _ = GraphCommandProcessor(graph, next_command_processor=_)
            command_processor = GraphCommandProcessor(graph, next_command_processor=_)

            command = GraphCommand(GraphCommandType.DEPLOY_TO_CUDA)
            command_processor(command)

        All three GraphCommandProcessor will then be called one by one

        Never attempt to use function like _(command) in above case,
        all responsibility chains should only be called by its head.

        Args:
            command (GraphCommand): An acceptable GraphCommand object
            if an improper GraphCommand is given, it will incurs ValueError at end.

        Raises:
            ValueError: raise when there is no suitable processor for your command.

        Returns:
            Any: processor will decide what is it result.
        """

        if not isinstance(command, GraphCommand):
            raise ValueError(
                "command should be an instance of GraphCommand,"
                f"however {type(command)} received yet."
            )

        if command.command_type in self._acceptable_command_types:
            return self.process(command)
        elif self._next_command_processor is not None:
            self._next_command_processor(command)
        else:
            raise ValueError(
                f"Command Type {command.command_type} is not acceptable in this graph, "
                "please make sure you have added proper command processor into "
                "processing chain.\n"
                f"For more information, you may refer to {self}"
            )

    @abstractmethod
    def process(self, command: GraphCommand) -> Any:
        """Subclass of GraphCommandProcessor must give an implementation of
        this function.

            Process received GraphCommand instance and give result(if there is any)

        Args:
            command (GraphCommand): input command object.

        Returns:
            Any: any result is fine.
        """
        raise NotImplementedError

    @property
    def graph(self) -> BaseGraph:
        return self._graph

    def __str__(self) -> str:
        return f"GraphCommandProcessor {self.__hash__}"



"""
Copyright Wenyi Tang 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

"""

import tempfile
from enum import IntEnum

import onnx
import pytest
import torch

from mppq.api import ENABLE_CUDA_KERNEL, quantize, register_platform


class TestPlatform(IntEnum):
    MY_PLATFORM = 1
    TEST1 = 999
    TEST2 = 1000
    TEST3 = 1001


register_platform(TestPlatform.MY_PLATFORM, {"allin": None}, {})


def test_register_platform():
    register_platform(TestPlatform.TEST1, {"allin": None}, {})
    with pytest.raises(KeyError):
        register_platform(TestPlatform.TEST1, {}, {})


@pytest.mark.skipif(not torch.cuda.is_available(), reason="CUDA is not available")
def test_simple_quantize(model):
    with ENABLE_CUDA_KERNEL(), tempfile.TemporaryDirectory() as tmpdir:
        quantize(model, f"{tmpdir}/test.onnx", TestPlatform.MY_PLATFORM)
        model = onnx.load(f"{tmpdir}/test.onnx")
        onnx.checker.check_model(model, True)



"""
Copyright Wenyi Tang 2025

:Author: Wenyi Tang
:Email: wenyitang@outlook.com

"""

# pylint: disable=missing-docstring

import onnx
from onnx.helper import make_graph, make_node, make_tensor_value_info

from mppq.api import load_onnx_graph
from mppq.dispatcher.scope import IgnoredScope, Subgraph


def _build_graph():
    r"""sigmoid    relu
         \       /
          \     /
           \   /
            add
             |
           split
            / \
           /   \
          /     \
       softmax  softmax
    """

    sigmoid = make_node("Sigmoid", ["input0"], ["sigmoid"], "my_name_is_sigmoid")
    relu = make_node("Relu", ["input1"], ["relu"], "my_name_is_relu")
    add = make_node("Add", ["sigmoid", "relu"], ["add"], "my_name_is_add")
    split = make_node(
        "Split",
        ["add"],
        ["softmax0", "softmax1"],
        "my_name_is_split",
        num_outputs=2,
        axis=1,
    )
    softmax0 = make_node("Softmax", ["softmax0"], ["output0"], "my_name_is_softmax0")
    softmax1 = make_node("Softmax", ["softmax1"], ["output1"], "my_name_is_softmax1")
    graph = make_graph(
        [sigmoid, relu, add, split, softmax0, softmax1],
        "test_graph",
        inputs=[
            make_tensor_value_info("input0", onnx.TensorProto.FLOAT, [1, 64]),
            make_tensor_value_info("input1", onnx.TensorProto.FLOAT, [1, 64]),
        ],
        outputs=[
            make_tensor_value_info("output0", onnx.TensorProto.FLOAT, [1, 32]),
            make_tensor_value_info("output1", onnx.TensorProto.FLOAT, [1, 32]),
        ],
    )
    model = onnx.helper.make_model(
        graph,
        ir_version=onnx.IR_VERSION_2023_5_5,
        opset_imports=[onnx.helper.make_operatorsetid("", 19)],
    )
    onnx.checker.check_model(model)
    return load_onnx_graph(model)


def test_ignore_by_operation_name():
    graph = _build_graph()
    scope = IgnoredScope(operations=["my_name_is_add"])
    disp = scope.dispatch(graph)
    assert len(disp) == 1
    assert "my_name_is_add" in disp


def test_ignore_by_operation_type():
    graph = _build_graph()
    scope = IgnoredScope(types=["Softmax"])
    disp = scope.dispatch(graph)
    assert len(disp) == 2
    assert "my_name_is_softmax0" in disp
    assert "my_name_is_softmax1" in disp


def test_ignore_by_fuzzy_name():
    graph = _build_graph()
    scope = IgnoredScope(operations=["*sigmoid"])
    assert "my_name_is_sigmoid" in scope.dispatch(graph)

    scope = IgnoredScope(operations=["my_name_is_softmax?"])
    disp = scope.dispatch(graph)
    assert len(disp) == 2
    assert "my_name_is_softmax0" in disp
    assert "my_name_is_softmax1" in disp

    scope = IgnoredScope(operations=[r"re://my_name_is_\w{3,4}"])
    disp = scope.dispatch(graph)
    assert len(disp) == 2
    assert "my_name_is_relu" in disp
    assert "my_name_is_add" in disp


def test_ignore_by_subgraph():
    graph = _build_graph()
    scope = IgnoredScope(
        subgraphs=[
            Subgraph(
                ["my_name_is_sigmoid", "my_name_is_relu"],
                ["my_name_is_softmax0"],
            )
        ]
    )
    disp = scope.dispatch(graph)
    assert len(disp) == 5
    assert "my_name_is_softmax1" not in disp
